#!/bin/bash

# ACCUMULATION OF NUMERATOR (ML) STATISTICS

bin/SERest \
  -p 1 -G HTK -D -T 3 \
  -S features.scp \
  -I numerator.mlf \
  -H GMM_ML/models \
  -M GMM_MMI \
  --nframe-outp-norm=TRUE \
  --outp-scale=6
    
# SERest is run in parallel mode (-p 1), thus the output is accumulator file
# 'SER1.acc' stored to directory 'GMM_MMI' (-M). Each line of feature list
# (script) file 'features.scp' (-S) is in the form:
# 
# logical_file_name=physical_file_name[seg_start,seg_end]{weight}
#  
# which follows the standard HTK specification except for the weight in braces.
# The weights allows to give different importance to individual utterances
# (and segments). For example, weighting segment by factor 2 would correspond to
# using the segment twice for training. However, the weights can be also decimal
# numbers. For the demonstration purpose, feature list 'features.scp' contains
# only one (segmented) utterance with approximately the same number of frames
# for each language. The exception is English with two utterances and about
# twice more feature frames. Therefore, the weight 0.5 is given to all English
# segments, which prevent MMI from learning the priors.
# 
# Logical file names are same for all segments corresponding to one language,
# i.e. the same label (file) will be used for all the segments. Standard HTK
# labels (-G HTK) are searched in MLF 'numerator.mlf' (-I).
# 
# Options --nframe-outp-norm=TRUE cause that acoustic scores (likelihoods) are
# normalized (divided) by number of frames in segment and option --outp-scale=6
# cause further scaling of the score by empirical factor 6. Both these options
# are only important for accumulation of denominator statistic (to increase
# confusion between alternative hypotheses) and, for this task, they do not
# affect computation of numerator (ML) accumulators. However, they scale
# reported numerator likelihood the same way as for denominator, which allows
# to computate meaningful MMI objective function (ratio of numerator and
# denominator likelihoods).


# ACCUMULATION OF DENOMINATOR STATISTICS

bin/SERest \
  -p 2 -G STK -D -T 3 \
  -S features.scp \
  -H GMM_ML/models \
  -M GMM_MMI \
  --recognet=denominator.net \
  --outp-scale=6 \
  --nframe-outp-norm=TRUE
  
# The output is accumulator file 'GMM_MMI/SER2.acc'. Statistics are accumulated
# using competing model given by network file 'denominator.net' (-G STK
# --recognet), which represents composed HMM having all the individual language
# (single state) HMMs connected in parallel. The network is similar to HTK
# standard lattice format (SLF), however, it can also contain node definitions
# referencing directly HMM's ('M=') rather than words ('W='). (Note that
# standard network consisting only of word nodes can be also used, however,
# than it would be necessary to specify dictionary (--sourcedict=dictfile)
# defining expansion of words to HMM's)


bin/SERest \
  -p 0 -D -T 3 -u "mv"\
  -H GMM_ML/models \
  -M GMM_MMI \
  --update-type=MMI \
  --EBW-constant-E=2 \
  GMM_MMI/SER1.acc GMM_MMI/SER2.acc   #1st num. and den. accumulator
# GMM_MMI/SER3.acc GMM_MMI/SER4.acc   #2nd num. and den. accumulator
# ...

# MMI update use both num. and den. accumulators. In this simple case,
# only one accumulator for numerator (SER1.acc) and one for denominator
# (SER2.acc) was generated. You will probably want to run generation of
# many accumulators in parallel on many machines. You need to split
# 'features.scp' file. For each part generate one numerator and one
# denominator accumulator and call the MMI update passing alternately
# numerator and denominator accumulators as program arguments
#
#
# Notes:
#
# The long options (--foo) are in fact configuration parameters. Any
# configuration parameter supported by STK can be either defined in config
# file (-C) or as long option. Characters '-' and '_' in option names
# are ignored, so they can be used to improve option name readability.
# Option names are also case insensitive. For example, to force STK tools
# to automatically compute MFCC derivatives the following options can be
# included --target-kind=MFCC_0_D_A --delta_window=3 --accWindow=3.
#
# STK also allows to specify non-trivial feature transforms, which, for
# example, allows for on-line computation of SDC parameters. In our
# NIST-LID'05, the features are 7-1-3-7 SDC + 7 MFCC (56 coefficients).
# The MMI training code above assume feature files containing such
# 56-dimensional features. If the feature files contained only 7-dimensional
# MFCC, the code would have to be modified as follows to compute the features
# on-line:
#
# 1. remove first line of file GMM_ML/models (features in a file are no more
#    56-dimensional)
# 2. Before each option '-H GMM_ML/models' insert '-H GMM_ML/sdc_transform',
#    which is macrofile specifying the transformation: MFCC -> MFCC+SDC
# 3. Because of the frame delay of SDC processing, 20 frames would be lost
#    from each speech segment. To compensate for this, options 
#    --start-frm-ext=10 --end-frm-ext=10 can be added, extending beginning
#    and end of each segment by 10 frames.
